/** 
 * This is rc102remote.c
 * 
 * A program to poll rc102remote device
 * and output event descriptors to stdout
 * 
 * Uses libhid. 
 *
 * Richard Kaye 2006-7
 * 
 * COPYRIGHT AND LICENCE
 * 
 * Developed by Richard Kaye <R.W.Kaye@bham.ac.uk>
 *
 * Incorporates code from test_libhid.c, part of libhid, 
 * which is (c) 2003-2005 by
 *   Martin F. Krafft <libhid@pobox.madduck.net>
 *   Charles Lepple <clepple@ghz.cc>
 *   Arnaud Quette <arnaud.quette@free.fr> && <arnaud.quette@mgeups.com>
 * and distributed under the terms of the GNU General Public License.
 *
 * Additions, modifications, etc, are copyright Richard Kaye
 * <R.W.Kaye@bham.ac.uk> and distributed under the terms 
 * of the GNU General Public License. See the file COPYING in the 
 * source distribution for more information.
 * 
 * THIS PACKAGE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
 * OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#include <hid.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>

/* function to identify hid device */
bool match_serial_number(struct usb_dev_handle* usbdev, 
			 void* custom, unsigned int len)
{
  bool ret;
  char* buffer = (char*)malloc(len);
  usb_get_string_simple(usbdev, 
			usb_device(usbdev)->descriptor.iSerialNumber,
			buffer, 
			len);
  ret = strncmp(buffer, (char*)custom, len) == 0;
  free(buffer);
  return ret;
}

/* log output file */
FILE *stdlog = NULL;

/* table of output commands */
char *outptab[14];

/* global parameters for the hid device */
HIDInterface* hid;
hid_return ret;

/* catch interrupt signal and close down gracefully */
void close_rc102(int sig)
{
  ret = hid_close(hid);
  if (ret != HID_RET_SUCCESS) {
    fprintf(stderr, "rc102remote: hid_close failed with return code %d\n", ret);
    exit( EXIT_FAILURE );
  }
  hid_delete_HIDInterface(&hid);
  ret = hid_cleanup();
  if (ret != HID_RET_SUCCESS) {
    fprintf(stderr, "rc102remote: hid_cleanup failed with return code %d\n", ret);
    exit( EXIT_FAILURE );
  }
  if (stdlog) fclose(stdlog);
  fflush(stderr);
  exit(0);
}

/* returns int value of a string */
int parseint(char *str)
{
    int ptr = 0;
    int i = 0;
    int digit;
    // fprintf(stderr,"scanning \"%s\"\n",str);
    while (*(str+ptr) == ' ') ptr++;
    while (*(str+ptr) != '\0' && 
	   *(str+ptr) !=' '   && 
	   *(str+ptr) !='\n'  && 
	   *(str+ptr) !='\v') {
      digit = *(str+ptr);
      // fprintf(stderr,"read \'%c\'\n", digit);
      digit -= '0';
      if (digit<0 || digit>9) {
        fprintf(stderr,"illegal format for integer: %s\n",str);
        exit( EXIT_FAILURE );
      }
      i = 10*i + digit;
      ptr ++;
    }
    // fprintf(stderr,"parsed %d\n", i);
    return i;
}

/* parses a 8 character hex string */
void parsepacket(char *str, unsigned char *packet)
{
    int ptr = 0;
    int count = 0;
    int digit;
    // fprintf(stderr,"scanning \"%s\"\n",str);
    while (*(str+ptr) == ' ') ptr++;
    for (count=0; count<8; count++) {
      digit = *(str+ptr);
      // fprintf(stderr,"read \'%c\'\n", digit);
      digit -= '0';
      if (digit<0) {
        fprintf(stderr,"rc102remote: illegal format for packet: %s (bad character = %c)\n",str,digit+'0');
        exit( EXIT_FAILURE );
      }
      if (digit>=10) {
	digit += '0' - 'a' + 10;
	if (digit<10 || digit>15) {
	  fprintf(stderr,"rc102remote: illegal format for packet: %s (bad character = %c)\n",str,digit+'a'-10);
	  exit( EXIT_FAILURE );
	}
      }      
      if (0==count%2) packet[count/2] = digit*16;
      else packet[count/2] = packet[count/2] + digit;
      ptr ++;
    }
    while (*(str+ptr) == ' ') ptr++;
    if (*(str+ptr) != '\0') {
      fprintf(stderr,"rc102remote: illegal format for packet: %s (bad character = %c)\n",str, *(str+ptr));
      exit( EXIT_FAILURE );
    }
}

/* main command-line parameters */
char *logfilename = NULL;              // file to log messages to, NULL for none
char *tabfilename = "rc102.table";     // file to get output commands
int append = 0;                        // whether to append to log file
int microsleepactive = 0;              // fine control on algorithm
int microsleepinactive = 0;            // fine control on algorithm
int settletime = -1;                   // fine control on algorithm
unsigned char mask[4] = { '\0', 0xf0, '\0', '\0' };
unsigned char control[4] = { '\0', 0xf0, '\0', '\0' };

/* main: read command line and start up HID, and initiate polling */
int main(int argc, char *argv[])
{
  /* option handling
   *   -lLOGFILE              name of logfile
   *   -a                     append to logfile
   *   -uMICROSLEEP           fine control on algorithm
   *   -fMICROSLEEPACTIVE     fine control on algorithm
   *   -bMICROSLEEPINACTIVE   fine control on algorithm
   *   -sSETTLETIME           fine control on algorithm
   *   -mMASK                 fine control on algorithm 
   *   -cCONTROL              fine control on algorithm
   *   -tTABLE                table of output lines
   */
  int opt;
  while( ( opt = getopt( argc, argv, "ahlt:n:u:f:b:s:m:c:" ) ) != EOF ) {
    switch( opt ) {
    case 'a': 
      append = 1; 
      break;
    case 'h':
      fprintf(stderr,"Usage: rc102remote [-l|-a] [-nLOGFILE] [-tDATAFILE] \n");
      // fprintf(stderr,"                   [-(u|f|b)MICROS] [-sSETTLETIME] [-mMASK] [-cCONTROL]\n");
      exit(0);
    case 'l': 
      logfilename = "rc102.log"; 
      break;
    case 'n': 
      logfilename = optarg; 
      break;
    case 't': 
      tabfilename = optarg; 
      break;
    case 'u': microsleepactive = microsleepinactive = parseint(optarg);
      break;
    case 'f': microsleepactive = parseint(optarg);
      break;
    case 'b': microsleepinactive = parseint(optarg);
      break;
    case 's': settletime = parseint(optarg);
      break;
    case 'm': parsepacket(optarg,mask);
      break;
    case 'c': parsepacket(optarg,control);
      break;
    default:
      fprintf(stderr, "Unknown option '%c'\n", opt );
      exit( EXIT_FAILURE );
    }
  }
  if (optind != argc) {
    fprintf(stderr,"Unknown option '%s'\n", argv[optind] );
    exit( EXIT_FAILURE );
  }
  if (logfilename) {
    if (append) stdlog = fopen(logfilename,"a");
    else stdlog = fopen(logfilename,"w");
    if (!stdlog) {
      fprintf(stderr,"rc102remote: cannot open log file '%s'\n", logfilename );
      exit( EXIT_FAILURE );
    }
  }
  if (stdlog) {
    fprintf(stdlog,"rc102remote: parameter microsleepinactive (-b) = %i\n",microsleepinactive);
    fprintf(stdlog,"rc102remote: parameter microsleepactive (-f) = %i\n",microsleepactive);
    fprintf(stdlog,"rc102remote: parameter mask (-m) = %.2x %.2x %.2x %.2x\n",mask[0], mask[1], mask[2], mask[3]);
    fprintf(stdlog,"rc102remote: parameter control (-c) = %.2x %.2x %.2x %.2x\n",control[0], control[1], control[2], control[3]);
  }

  /* Read 14 lines from outptab corresponding to keys 0-13 */
  if (stdlog) fprintf(stdlog, "rc102remote: reading data file %s\n", tabfilename);
  FILE *foutptab = fopen(tabfilename,"r");
  if (!foutptab) {
    fprintf(stderr,"rc102remote: cannot open data file '%s'\n", tabfilename );
    exit( EXIT_FAILURE );
  }
  int linen = 0;
  char strspace[4096]; // TO DO: what if this overflows?
  int spavail = 4096;
  char *p, *q;
  p = strspace;
  while (linen<14 && spavail>0) {
    // fprintf(stderr, "line  %i: ", linen);
    fgets ( p, spavail, foutptab );
    if (feof(foutptab)) break;
    // fprintf(stderr, "%s", p);
    // replace ";" with "\n" in the string just read
    q = p; while (*q!='\0') { if (*q==';') *q='\n'; q++; }
    outptab[linen] = p;
    spavail -= strlen(p)+1;
    p = p + strlen(p)+1;
    // fprintf(stderr, "line %i: %s", linen, outptab[linen]);
    linen++;
  }
  while (linen < 14) {
    outptab[linen] = "";
    // fprintf(stderr, "line %i: %s", outptab[linen]);
    linen++;
  }
  fclose(foutptab);

  /* Now open device... the following doc is from libhid
   * 
   ** The third member of the HIDInterfaceMatcher is a function pointer, and
   ** the forth member will be passed to it on invocation, together with the
   ** USB device handle. The fifth member holds the length of the buffer
   ** passed. See above. This can be used to do custom selection e.g. if you
   ** have multiple identical devices which differ in the serial number.
   **
   **   char const* const serial = "01518";
   **   HIDInterfaceMatcher matcher = {
   **     0x147a,                      // vendor ID
   **     0xe019,                      // product ID
   **     match_serial_number,         // custom matcher function pointer
   **     (void*)serial,               // custom matching data
   **     strlen(serial)+1             // length of custom data
   **   };
   **
   ** Third member may be NULL, then the match will only be on vendor and 
   ** product ID. 
   */
  HIDInterfaceMatcher matcher = { 0x147a, 0xe019, NULL, NULL, 0 };

  /* see include/hid.h (typedef enum HIDDebugLevel_t) 
   * for possible values; HID_DEBUG_ALL is far too much...
   */
  hid_set_debug( HID_DEBUG_ERRORS | HID_DEBUG_WARNINGS );
  if (stdlog) hid_set_debug_stream(stdlog);
  else hid_set_debug_stream(stderr);

  /* This is passed directly to libusb */
  hid_set_usb_debug(0); // presumably meaning NONE
  
  /* initialize the device */
  ret = hid_init();
  if (ret != HID_RET_SUCCESS) {
    fprintf(stderr, "rc102remote: hid_init failed with return code %d\n", ret);
    exit( EXIT_FAILURE );
  }

  /* get interface */
  hid = hid_new_HIDInterface();
  if (hid == 0) {
    fprintf(stderr, "rc102remote: hid_new_HIDInterface() failed, out of memory?\n");
    exit( EXIT_FAILURE );
  }

  /* How to detach a device from the kernel HID driver:
   * 
   ** The hid.o or usbhid.ko kernel modules claim a HID device on insertion,
   ** usually. To be able to use it with libhid, you need to blacklist the
   ** device (which requires a kernel recompilation), or simply tell libhid to
   ** detach it for you. hid_open just opens the device, hid_force_open will
   ** try n times to detach the device before failing.In the following, n == 3.
   **
   ** To open the HID, you need permission to the file in the /proc usbfs
   ** (which must be mounted -- most distros do that by default):
   **   mount -t usbfs none /proc/bus/usb
   ** You can use hotplug to automatically give permissions to the device on
   ** connection. Please see
   **   http://cvs.ailab.ch/cgi-bin/viewcvs.cgi/external/libphidgets/hotplug/
   ** for an example. Try NOT to work as root!
   */
  ret = hid_force_open(hid, 0, &matcher, 3);
  if (ret != HID_RET_SUCCESS) {
    fprintf(stderr, "rc102remote: hid_force_open failed with return code %d\n", ret);
    exit( EXIT_FAILURE );
  }

  /* Now get identification data and write to stdlog */
  if (stdlog) ret = hid_write_identification(stdlog, hid);
  if (ret != HID_RET_SUCCESS) {
    fprintf(stderr, "rc102remote: hid_write_identification failed with return code %d\n", ret);
    exit( EXIT_FAILURE );
  }
  /* should see something like:
   * 
   * device identification of HIDInterface 001/004[0]:
   *   dev_handle:    0x0804b1f0
   *   device:        0x08056468
   *   location:      001/004
   *   manufacturer:  Formosa Industrial Computin
   *   product:       Formosa RC102-809 USB Remot
   */

  /* Now get usage data and write to stdlog */
  if (stdlog) ret = hid_dump_tree(stdlog, hid);
  if (ret != HID_RET_SUCCESS) {
    fprintf(stderr, "rc102remote: hid_dump_tree failed with return code %d\n", ret);
    exit( EXIT_FAILURE );
  }
  /*
   * should get something like:
   * 
   * parse tree of HIDInterface 001/004[0]:
   *   path: 0x000c0001.0x000c0001; type: 0x80
   *   path: 0x000c0001.0x00000000; type: 0x80
   *   path: 0x000c0001.0x00000000; type: 0x80
   *   path: 0x000c0001.0x00000000; type: 0x80
   */

  /* How to write to and read from a device. The following is from libhid.
   *
   ** Writing to a device requires the HID usage path, a buffer, and the length
   ** of the latter. You must also know the exact length of a packet expected
   ** by the device, and the protocol to speak over HID.
   **
   ** libhid uses the MGE hidparser, which parses the HID usage tree and places
   ** the available usages at its leafs (leaves?). The path information can be
   ** read from the `lsusb -vvv` output, or by inspecting the output of
   ** hid_dump_tree. In the output, 0x80 denotes an input endoint (sent by the
   ** device), and 0x90 an output endpoint (sent to the device). These are then
   ** used to communicate with the device.
   **
   ** In the example of the Phidgets QuadServoController (www.phidgets.com, and
   ** libphidgets.alioth.debian.org), the following two paths identify the
   ** input and output descriptors respectively.
   **
   ** This is derived from the output of `lsusb -d 0x147a:0xe019 -vvv` as
   ** follows. You need to run `libhid_detach_device 147a:e019` before lsusb
   ** will output this info:
   **
   **   Bus 001 Device 028: ID 147a:e019 Formosa Industrial Computing, Inc.
   **   Device Descriptor:
   **     [...]
   **     Configuration Descriptor:
   **       [...]
   **       Interface Descriptor:
   **         [...]
   **         iInterface
   **           HID Device Descriptor:
   **           [...]
   **              Report Descriptor: // (length is 20)
   **                Item(Global): Usage Page, data= [ 0x0c ] 12
   **                                Consumer
   **                Item(Local ): Usage, data= [ 0x01 ] 1
   **                                Consumer Control
   **                Item(Main  ): Collection, data= [ 0x00 ] 0
   **                                Physical
   **                Item(Local ): Usage, data= [ 0x01 ] 1
   **                                Consumer Control
   **                Item(Global): Logical Minimum, data= [ 0x00 ] 0
   **                Item(Global): Logical Maximum, data= [ 0xff 0x00 ] 255
   **                Item(Global): Report Size, data= [ 0x08 ] 8
   **                Item(Global): Report Count, data= [ 0x04 ] 4
   **                Item(Main  ): Input, data= [ 0x06 ] 6
   **                                Data Variable Relative No_Wrap Linear
   **                                Preferred_State No_Null_Position Non_Volatile Bitfield
   **                Item(Main  ): End Collection, data=none
   *
   * I don't understand any of this!  The number 0x000c0001 
   * (page=0x000c, usage=0x0001 is clearly important. 
   *
   * Fortunately either my guess was lucky or this device is very simple...
   */
  
  /*
   *  // Example code for writing to device to send 6 bytes:
   *  // we don't actually need to write for this particular device
   *
   *  unsigned char const PATHLEN = 3;
   *  int const PATH_IN[] = { 0x000c0001, 0x000c0001, 0x000c0001 };
   *  int const PATH_OUT[] = { 0x000c0001, 0x000c0001, 0x000c0001 };
   *
   *  unsigned char const SEND_PACKET_LEN = 6;
   *  // fill an example packet:
   *  char const PACKET[] = { 0x01, 0x02, 0x03, 0x04, 0x05, 0x06 };
   *  //
   *  ret = hid_set_output_report(hid,PATH_IN,PATHLEN,PACKET,SEND_PACKET_LEN);
   *  if (ret != HID_RET_SUCCESS) {
   *      fprintf(stderr,"rc102remote: hid_set_output_report failed with return code %d\n",ret);
   *  }
   */

  /* allow the OS to interrupt us so we can close nicely... */
  signal(SIGINT, &close_rc102);

  /* HID device and other code is now set up.  
   * We run main polling function, which writes to stdout. */
  poll_HID_device();

  /* if we get this far we close and exit */
  close_rc102(0);
}

/* This is the function that polls the device */
int poll_HID_device()
{
  /*
   * This device returns 4-byte sequences (I think) 
   */
  unsigned char const PATHLEN = 3;
  int const PATH_OUT[] = { 0x000c0001, 0x000c0001, 0x000c0001 };

  unsigned char packet[4]; // for the packet just read
  unsigned char lastpk[4] = { '\0', '\0', '\0', '\0' }; // last packet
  int settle = settletime / (1+microsleepactive);
  // fprintf(stderr,"using settle = %i, settletime = %i, microsleepactive = %i\n", 
  //         settle, settletime, microsleepactive);
  int count = 0;      // number of consecutive identical values found
  int totalcount = 0; // total number of frames read
  int eventcount = 0; // total number of events found
  int match = 0;
  int j=0;

  /* continuously poll device until process is killed */
  while (1) { 
    count = totalcount = 0;
    
    /* loop, reading data, until input is constant for settle-time
     * or (TO DO) a handshake is detected using mask and control */
    while (1) {
      
      usleep(microsleepactive);
      ret = hid_get_input_report(hid, PATH_OUT, PATHLEN, packet, 4);
      if (ret != HID_RET_SUCCESS) {
	fprintf(stderr, "rc102remote: hid_get_input_report failed with return code %d\n", ret);
      }
      totalcount++;
      match = 1;
      for (j=0; j<4; j++) {
	if ((packet[j] & mask[j])!=control[j]) { match=0; break; }	
      }
      if (match) {
	for (j=0; j<4; j++) lastpk[j]=packet[j];
	break;
      }
      
      if ( packet[0]!=lastpk[0] || 
	   packet[1]!=lastpk[1] || 
	   packet[2]!=lastpk[2] || 
	   packet[3]!=lastpk[3] ) {
	count = 0;
	lastpk[0]=packet[0];
	lastpk[1]=packet[1];
	lastpk[2]=packet[2];
	lastpk[3]=packet[3];
	continue;
      }

      count++;
      if (count < 0) count=0;
      if (settle>=0 && count>=settle) break;
      
    }

    eventcount++;
    // fprintf(stdlog, "%.2x%.2x%.2x%.2x %i %i\n", packet[0], packet[1], 
    //         packet[2], packet[3], count*microsleepactive, eventcount);
    printf("%s",outptab[packet[0]]);
    fflush(stdout);

    // now wait until something new happens
    while (1) {
      
      usleep(microsleepinactive);
      ret = hid_get_input_report(hid, PATH_OUT, PATHLEN, packet, 4);
      if (ret != HID_RET_SUCCESS) {
	fprintf(stderr, "rc102remote: hid_get_input_report failed with return code %d\n", ret);
      }
      totalcount++;
      
      if ( packet[0]!=lastpk[0] || 
	   packet[1]!=lastpk[1] || 
	   packet[2]!=lastpk[2] || 
	   packet[3]!=lastpk[3] ) {
	count = 0;
	lastpk[0]=packet[0];
	lastpk[1]=packet[1];
	lastpk[2]=packet[2];
	lastpk[3]=packet[3];
	break;
      }
    }

  }
  return 0;
}
